"""Mysql database service"""
import logging
from sqlalchemy import create_engine, text
from sqlalchemy.exc import OperationalError
import pandas as pd

from pydisadm.services.common import (
    CREATE_TABLE_ADM, CREATE_TABLE_ADM_HISTORY, CREATE_TABLE_MAP
)
from pydisadm.services.database import Database

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

logger.addHandler(ch)

POOL_RECYCLE_SECONDS=60*10
RETRIES=3

def query_retry(retries, exceptions):
    """Retry queries in case of failure."""
    def _inner_func(func):
        def _retry_wrapper(*args, **kwargs):
            attempt = 0
            while attempt < retries:
                try:
                    return func(*args, **kwargs)
                except exceptions as exc:
                    logger.error('error %s on attempt %d of %d', exc, attempt + 1, retries)
                    attempt += 1

            return func(*args, **kwargs)
        return _retry_wrapper
    return _inner_func

class DatabaseMysql(Database):
    """Mysql database service implementation"""

    def __init__(self, conn_string):
        self.engine = create_engine(
            f'mysql+mysqldb://{conn_string}',
            pool_size=5,
            pool_pre_ping=True,
            pool_recycle=POOL_RECYCLE_SECONDS,
            echo_pool=True
        )
        self.setup()

    def setup(self):
        """Setup database schema"""
        with self.engine.connect() as conn:
            conn.execute(text(CREATE_TABLE_ADM))
            conn.execute(text(CREATE_TABLE_ADM_HISTORY))
            conn.execute(text(CREATE_TABLE_MAP))
            conn.commit()

    @query_retry(retries=RETRIES, exceptions=(OperationalError,))
    def insert_systems(self, systems):
        """Insert system ADM records"""
        with self.engine.connect() as conn:
            try:
                systems.to_sql('adm', conn, index=False, if_exists='append')
                systems.to_sql('adm_history', conn, index=False, if_exists='append')
            except OperationalError as error:
                logger.error('insert_systems error', exc_info=error)
                conn.rollback()
                raise error

    @query_retry(retries=RETRIES, exceptions=(OperationalError,))
    def insert_map_data(self, map_data):
        """Insert map data"""
        with self.engine.connect() as conn:
            try:
                map_data.to_sql('map', conn, index=True, if_exists='replace')
            except OperationalError as error:
                logger.error('insert_map_data error', exc_info=error)
                conn.rollback()
                raise error

    @query_retry(retries=RETRIES, exceptions=(OperationalError,))
    def select_system_with_name(self, system_name) -> pd.DataFrame:
        """Select system matching name"""
        with self.engine.connect() as conn:
            try:
                system = pd.read_sql_query(
                    text("SELECT * FROM map WHERE solarSystemName = :system_name"),
                    conn, params={ 'system_name': system_name })
            except OperationalError as error:
                logger.error('select_system_with_name error', exc_info=error)
                conn.rollback()
                raise error

        return system

    @query_retry(retries=RETRIES, exceptions=(OperationalError,))
    def select_most_recent_row(self) -> pd.DataFrame:
        """Select the most recent ADM record"""
        with self.engine.connect() as conn:
            try:
                most_recent = pd.read_sql_query(text("""
                    SELECT created_at FROM adm ORDER BY created_at DESC LIMIT 1
                """), conn)
            except OperationalError as error:
                logger.error('select_most_recent_row error', exc_info=error)
                conn.rollback()
                raise error

        return most_recent

    @query_retry(retries=RETRIES, exceptions=(OperationalError,))
    def select_systems(self) -> pd.DataFrame:
        """Select most recent record of all systems"""

        with self.engine.connect() as conn:
            try:
                systems = pd.read_sql_query(text("""
                    SELECT t1.*, t2.solarSystemName, t2.constellationName, t2.regionName FROM adm t1 
                    LEFT JOIN map t2 ON t1.system_id = t2.solarSystemID 
                    WHERE t1.created_at = (SELECT MAX(t3.created_at) FROM adm t3 WHERE t3.system_id = t1.system_id);
                """), conn)
            except OperationalError as error:
                logger.error('select_systems error', exc_info=error)
                conn.rollback()
                raise error

        return systems

    @query_retry(retries=RETRIES, exceptions=(OperationalError,))
    def select_system_by_name(self, name) -> pd.DataFrame:
        """Select systems by name"""
        sql = text("""
            SELECT map.solarSystemName system_name, adm, tier, created_at FROM adm
            INNER JOIN map ON map.solarSystemID = adm.system_id
            WHERE map.solarSystemName=:system_name OR
                map.constellationName=:constellation_name OR
                map.regionName=:region_name ORDER BY created_at
        """)
        with self.engine.connect() as conn:
            try:
                system = pd.read_sql_query(sql, conn, params={ 'system_name': name,
                                        'constellation_name': name, 'region_name': name })
            except OperationalError as error:
                logger.error('select_system_by_name error', exc_info=error)
                conn.rollback()
                raise error

        return system

    @query_retry(retries=RETRIES, exceptions=(OperationalError,))
    def select_system_history(self, system, limit) -> pd.DataFrame:
        """Select history of a single system"""

        with self.engine.connect() as conn:
            try:
                system_history = pd.read_sql_query(text("""
                    SELECT system_id, adm, tier, created_at FROM adm
                    INNER JOIN map ON map.solarSystemID = systems.system_id
                    WHERE map.solarSystemName=:system_name ORDER BY created_at DESC LIMIT :limit
                """), conn, params={'system_name': system, 'limit': limit })
            except OperationalError as error:
                logger.error('select_system_history error', exc_info=error)
                conn.rollback()
                raise error

        return system_history

    @query_retry(retries=RETRIES, exceptions=(OperationalError,))
    def delete_system_rows(self, days_old):
        """Delete system rows older than days_old"""

        with self.engine.connect() as conn:
            try:
                conn.execute(text("""
                    DELETE FROM adm WHERE created_at < DATE(NOW()-INTERVAL :days DAY)
                """), parameters={ 'days': days_old })

                conn.commit()
            except OperationalError as error:
                logger.error('delete_system_rows error', exc_info=error)
                conn.rollback()
                raise error



"""Datasets"""
from importlib import resources

def get_solar_systems():
    """Retrieve solar systems dataset path"""
    with resources.path('pydisadm.data', 'mapSolarSystems.csv') as path:
        file_path = path
    return file_path

def get_constellations():
    """Retrieve constellations dataset path"""
    with resources.path('pydisadm.data', 'mapConstellations.csv') as path:
        file_path = path
    return file_path

def get_regions():
    """Retrieve regions dataset path"""
    with resources.path('pydisadm.data', 'mapRegions.csv') as path:
        file_path = path
    return file_path



"""Application configuration"""

import os

class Configuration:
    """App configuration variables"""

    DB_SERVICES = ['sqlite', 'mysql']

    def __init__(self):

        self.discord = {
            'token': os.getenv('DISCORD_TOKEN'),
            'channel': os.getenv('DISCORD_CHANNEL'),
            'app_id': os.getenv('DISCORD_APP_ID'),
            'guild_id': os.getenv('DISCORD_GUILD_ID')
        }

        self.database = {
            'service': os.getenv('DB_SERVICE', 'sqlite'),
            'connection_string': os.getenv('DB_CONNECTION_STRING'),
            'keep_adm_days': os.getenv('DB_KEEP_ADM_DAYS', '7')
        }

        if self.database['service'] not in Configuration.DB_SERVICES:
            raise ValueError(
                f'[configuration] unsupported `DB_SERVICE`: {self.database["service"]}'
            )

        self.alliance = {
            'id': os.getenv('ALLIANCE_ID'),
            'ignore_tcu': os.getenv('ALLIANCE_IGNORE_TCU')
        }

        keep_adm_days = self.database['keep_adm_days']
        if keep_adm_days is not None:
            try:
                keep_adm_days = int(keep_adm_days)
            except ValueError as error:
                raise ValueError(
                    '[configuration] invalid `DB_KEEP_ADM_DAYS` value',
                    keep_adm_days) from error

        self.database['keep_adm_days'] = keep_adm_days

        alliance_id = self.alliance['id']
        if alliance_id is not None:
            try:
                alliance_id = int(alliance_id)
            except ValueError as error:
                raise ValueError(
                    '[configuration] invalid `ALLIANCE_ID` value',
                    alliance_id) from error

        self.alliance['id'] = alliance_id

    def pretty_print(self):
        """Create printable string from configuration"""
        return (f'{self.__class__.__name__}' +
            f'(discord_token={self.discord["token"]}, ' +
            f'discord_channel={self.discord["channel"]}, ' +
            f'discord_app_id={self.discord["app_id"]}, ' +
            f'discord_guild_id={self.discord["guild_id"]}, ' +
            f'alliance_id={self.alliance["id"]}, ' +
            f'alliance_ignore_tcu={self.alliance["ignore_tcu"]}, ' +
            f'db_keep_adm_days={self.database["keep_adm_days"]}, ' +
            f'db_service={self.database["service"]}, ' +
            f'db_connection_string={self.database["connection_string"]})')

    def __str__(self):
        return self.pretty_print()



"""Controller implementation for ADM operations"""
import os
import numpy as np
import pandas as pd
from tabulate import tabulate

from pydisadm.configuration import Configuration
from pydisadm.services.database_sqlite import DatabaseSqlite
from pydisadm.services.esi import sovereignty_structures
from pydisadm.utils.adm_utils import adm_from_index
from pydisadm.utils.plot_utils import plot_adm_history_of_systems, plot_save_to_file


class AdmController:
    """Controller implementation for ADM operations"""

    def __init__(self, configuration: Configuration, database: DatabaseSqlite):
        self.configuration = configuration
        self.database = database

    def generate_tier_list(self):
        """Generate a system tier list based on tier values"""
        system_adms = self.database.select_systems()

        system_adms.sort_values(by='adm', inplace=True, ascending=True)

        s_tier = system_adms.loc[system_adms['tier'] == 's_tier']
        a_tier = system_adms.loc[system_adms['tier'] == 'a_tier']
        b_tier = system_adms.loc[system_adms['tier'] == 'b_tier']
        c_tier = system_adms.loc[system_adms['tier'] == 'c_tier']
        d_tier = system_adms.loc[system_adms['tier'] == 'd_tier']

        sorted_summary = [
            {'Tier': 'D', 'Systems': '\n'.join(d_tier['solarSystemName']),
             'ADM': '\n'.join(d_tier['adm'].map(str)),
             'Constellation': '\n'.join(d_tier['constellationName']),
             'Region': '\n'.join(d_tier['regionName'])},
            {'Tier': 'C', 'Systems': '\n'.join(c_tier['solarSystemName']),
             'ADM': '\n'.join(c_tier['adm'].map(str)),
             'Constellation': '\n'.join(c_tier['constellationName']),
             'Region': '\n'.join(c_tier['regionName'])},
            {'Tier': 'B', 'Systems': '\n'.join(b_tier['solarSystemName']),
             'ADM': '\n'.join(b_tier['adm'].map(str)),
             'Constellation': '\n'.join(b_tier['constellationName']),
             'Region': '\n'.join(b_tier['regionName'])},
            {'Tier': 'A', 'Systems': '\n'.join(a_tier['solarSystemName']),
             'ADM': '\n'.join(a_tier['adm'].map(str)),
             'Constellation': '\n'.join(a_tier['constellationName']),
             'Region': '\n'.join(a_tier['regionName'])},
            {'Tier': 'S', 'Systems': '\n'.join(s_tier['solarSystemName']),
             'ADM': '\n'.join(s_tier['adm'].map(str)),
             'Constellation': '\n'.join(s_tier['constellationName']),
             'Region': '\n'.join(s_tier['regionName'])}
        ]

        table = tabulate(sorted_summary, showindex=False, headers='keys',
                         tablefmt='fancy_grid', numalign='left', stralign='center')

        generated_at = self.database.select_most_recent_row()

        if not generated_at.empty:
            generated_time = str(generated_at['created_at'][0])
        else:
            generated_time = None

        return (table, generated_time)

    def create_history_graph(self, name, file_name):
        """Generate an ADM history graph and save to file"""
        systems_by_name = self.database.select_system_by_name(name)

        if plot_adm_history_of_systems(systems_by_name):
            plot_save_to_file('ADM History', file_name)

        return os.path.isfile(file_name)

    def create_spreadsheet(self, file_name):
        """Generate a spreadsheet of ADM data and save to file"""
        system_adms = self.database.select_systems()
        system_adms.sort_values(by='adm', inplace=True, ascending=False)

        csv_columns = [
            'system_id',
            'adm',
            'tier',
            'created_at',
            'solarSystemName',
            'constellationName',
            'regionName'
        ]

        system_adms.to_csv(file_name, index=False, columns=csv_columns)

        return os.path.isfile(file_name)

    def tier_list(self, systems):
        """Assign tier column based on system ADM"""
        conditions = [
            (systems['adm'] >= 6.0),
            (systems['adm'] >= 5.0) & (systems['adm'] < 6.0),
            (systems['adm'] >= 4.0) & (systems['adm'] < 5.0),
            (systems['adm'] >= 3.5) & (systems['adm'] < 4.0),
            (systems['adm'] < 3.5)
        ]

        values = [
            's_tier',
            'a_tier',
            'b_tier',
            'c_tier',
            'd_tier'
        ]

        systems['tier'] = np.select(conditions, values)

        return systems

    def create_system_adm_from_index(self, system, military: int, industrial: int, strategic: int):
        """Create a DataFrame for system ADM data from index values"""
        adm = adm_from_index(military, industrial, strategic)

        return self.create_system_adm(system, adm)


    def create_system_adm(self, system, adm):
        """Create a DataFrame for system ADM data"""
        system_adm = pd.DataFrame(columns=['system_id', 'adm'])
        system_adm['system_id'] = system['solarSystemID']
        system_adm['adm'] = [adm]

        self.tier_list(system_adm)

        return system_adm

    def get_alliance_structures(self, structures, alliance_id, ignore_tcu):
        """Retrieve a list of alliance sovreignty structures"""
        structure_data = pd.DataFrame.from_dict(structures)

        if ignore_tcu:
            structure_data = structure_data[structure_data['structure_type_id'] != 32226]

        drop_columns = [
            'structure_id',
            'structure_type_id',
            'vulnerable_end_time',
            'vulnerable_start_time',
        ]

        structure_data.drop(columns=drop_columns, inplace=True)

        return structure_data[structure_data['alliance_id'] == alliance_id]

    def update_system_adm(self, system_name: str, adm: float) -> bool:
        """Update ADM for system with name"""
        system = self.database.select_system_with_name(system_name)

        if system.empty:
            return False

        insert_systems = self.create_system_adm(system, adm)

        self.database.insert_systems(insert_systems)

        return True

    def update_system_adm_from_index(self,
                                     system_name: str,
                                     military: int,
                                     industrial: int,
                                     strategic: int) -> bool:
        """Update ADM for system with name from index values"""
        system = self.database.select_system_with_name(system_name)

        if system.empty:
            return (False, 0)

        insert_systems = self.create_system_adm_from_index(
            system, military, industrial, strategic)

        self.database.insert_systems(insert_systems)

        return (True, insert_systems['adm'][0])

    def get_system_adms(self, alliance_id, ignore_tcu):
        """Retrieve ADM for systems controlled by alliance"""
        structures = sovereignty_structures()
        system_adms = self.get_alliance_structures(structures, alliance_id, ignore_tcu)

        system_adms.drop_duplicates(inplace=True)

        drop_columns = ['alliance_id']
        system_adms.drop(columns=drop_columns, inplace=True)

        rename_columns = {'solar_system_id': 'system_id',
                          'vulnerability_occupancy_level': 'adm'}
        system_adms.rename(columns=rename_columns, inplace=True)

        self.tier_list(system_adms)

        return system_adms

    def get_recommended_system(self):
        """Retrieve recommended system to raise ADM"""
        systems = self.database.select_systems()
        systems.sort_values(by='adm', inplace=True, ascending=True)

        if not systems.empty:
            return systems.iloc[0]

        return None

    def update_adm_data(self):
        """Update ADM data"""
        alliance_id = self.configuration.alliance['id']
        ignore_tcu = self.configuration.alliance['ignore_tcu']
        system_adms = self.get_system_adms(alliance_id, ignore_tcu)

        self.database.insert_systems(system_adms)

    def purge_adm_records(self, days_old: int):
        """Purge adm records more than days_old days old"""

        self.database.delete_system_rows(days_old)

    def write_file(self, file_name, content) -> bool:
        """Write content to file"""
        with open(file_name, 'w', encoding='UTF-8') as file:
            file.write(content)

        return os.path.isfile(file_name)

    def delete_file(self, file_name):
        """Delete file"""
        os.remove(file_name)



"""Modal dialog for updating ADM"""
import discord
from discord import ui

from pydisadm.controller.adm_controller import AdmController

class UpdateAdmModal(ui.Modal):
    """Modal dialog for updating ADM"""
    def __init__(self, system_name: str, controller: AdmController):
        super().__init__(title=f"Update {system_name} ADM")
        self.system_name = system_name
        self.controller = controller

    strategic_index = ui.TextInput(label='Strategic Index')
    military_index = ui.TextInput(label='Military Index')
    industry_index = ui.TextInput(label='Industry Index')

    # pylint: disable=arguments-differ
    async def on_submit(self, interaction: discord.Interaction):
        """on_submit callback"""

        try:
            strategic = int(self.strategic_index.value)
        except ValueError:
            await interaction.response.send_message(
                f'Strategic index not a number: {self.strategic_index}',
                ephemeral=True)
            return

        try:
            military = int(self.military_index.value)
        except ValueError:
            await interaction.response.send_message(
                f'Military index not a number: {self.military_index}',
                ephemeral=True)
            return

        try:
            industrial = int(self.industry_index.value)
        except ValueError:
            await interaction.response.send_message(
                f'Industrial index not a number: {self.industry_index}',
                ephemeral=True)
            return

        if not 0 <= strategic <= 5:
            await interaction.response.send_message(
                f'Strategic intex out of range: {self.strategic_index}',
                ephemeral=True)
            return

        if not 0 <= military <= 5:
            await interaction.response.send_message(
                f'Military index out of range: {self.military_index}',
                ephemeral=True)
            return

        if not 0 <= industrial <= 5:
            await interaction.response.send_message(
                f'Industrial index out of range: {self.industry_index}',
                ephemeral=True)
            return

        await interaction.response.defer()

        (result, adm) = self.controller.update_system_adm_from_index(
            self.system_name, military, industrial, strategic)

        if not result:
            await interaction.followup.send(
                f'Unable to update ADM for system: {self.system_name}',
                ephemeral=True)
        else:
            await interaction.followup.send(f'Updated ADM: **{self.system_name}** -> **{adm}**')



"""ESI API"""
import logging
import requests

logger = logging.getLogger('esi_api')

BASE_ENDPOINT="https://esi.evetech.net/latest"

def safe_get(url):
    """Safe get request"""
    try:
        response = requests.get(url, timeout=5000)
    except requests.RequestException as request_exception:
        logger.error(request_exception)
        return None

    return response


def safe_post(url, json):
    """Safe post request"""
    try:
        response = requests.post(url, json=json, timeout=5000)
    except requests.RequestException as request_exception:
        logger.error(request_exception)
        return None

    return response


def universe_names(ids):
    """Get names from IDs"""
    url = f"{BASE_ENDPOINT}/universe/names/?datasource=tranquility"
    return safe_post(url, ids).json()


def universe_system(system_id):
    """Get system by ID"""
    url = f"{BASE_ENDPOINT}/universe/systems/{system_id}?datasource=tranquility"
    return safe_get(url).json()


def universe_constellation(constellation_id):
    """Get constellation by ID"""
    url = f"{BASE_ENDPOINT}/universe/constellations/{constellation_id}?datasource=tranquility"
    return safe_get(url).json()


def universe_region(region_id):
    """Get region by ID"""
    url = f"{BASE_ENDPOINT}/universe/regions/{region_id}?datasource=tranquility"
    return safe_get(url).json()


def sovereignty_structures():
    """Get sovereignty structures"""
    url = f"{BASE_ENDPOINT}/sovereignty/structures/?datasource=tranquility"
    return safe_get(url).json()



"""Datetime utility functions"""
from datetime import datetime
import time
from dateutil.tz import tzutc, tzlocal

def convert_to_local_timestamp(date: str) -> int:
    """Convert to unix timestamp from date string"""
    generated_utc_date = datetime.fromisoformat(date)
    generated_utc_date = generated_utc_date.replace(tzinfo=tzutc())

    generated_local_date = generated_utc_date.astimezone(tzlocal())

    return int(time.mktime(generated_local_date.timetuple()))



"""Database service"""
from abc import abstractmethod

import pandas as pd

class Database:
    """Database service"""

    @abstractmethod
    def setup(self):
        """Setup database schema"""

    @abstractmethod
    def insert_systems(self, systems):
        """Insert system ADM records"""

    @abstractmethod
    def insert_map_data(self, map_data):
        """Insert map data"""

    @abstractmethod
    def select_system_with_name(self, system_name) -> pd.DataFrame:
        """Select system matching name"""

    @abstractmethod
    def select_most_recent_row(self) -> pd.DataFrame:
        """Select the most recent ADM record"""

    @abstractmethod
    def select_systems(self) -> pd.DataFrame:
        """Select most recent record of all systems"""

    @abstractmethod
    def select_system_by_name(self, name) -> pd.DataFrame:
        """Select systems by name"""

    @abstractmethod
    def select_system_history(self, system, limit) -> pd.DataFrame:
        """Select history of a single system"""

    @abstractmethod
    def delete_system_rows(self, days_old):
        """Delete system rows older than days_old"""



"""Factory to create database implementation"""

from pydisadm.services.database_mysql import DatabaseMysql
from pydisadm.services.database_sqlite import DatabaseSqlite

def create_database(configuration):
    """Create database implementation based on configuration."""

    db_service = configuration.database['service']
    connection_string = configuration.database['connection_string']
    if db_service == 'sqlite':
        database = DatabaseSqlite(connection_string)
    elif db_service == 'mysql':
        database = DatabaseMysql(connection_string)

    return database



"""Cog providing ADM related commands"""

import discord
from discord.ext import commands
from discord import app_commands
from discord.ext.commands import Context
import pydisadm
from pydisadm.bot.update_adm_modal import UpdateAdmModal

from pydisadm.bot.utils import check_allowed_channel, text_channels_with_send_permission
from pydisadm.configuration import Configuration
from pydisadm.controller.adm_controller import AdmController
from pydisadm.utils.datetime_utils import convert_to_local_timestamp

class Adm(commands.GroupCog):
    """Cog providing ADM related commands"""
    def __init__(self, bot, configuration: Configuration, controller: AdmController):
        self.bot = bot
        self.configuration = configuration
        self.controller = controller

    async def send_tier_list_summary(self, interaction: discord.Interaction):
        """Send a summary of ADMs organized as a tier list"""
        await interaction.response.defer(thinking=True)

        (tier_list, generated_at) = self.controller.generate_tier_list()
        if generated_at is None:
            await interaction.followup.send('😥 No ADM data...')
            return

        timestamp = convert_to_local_timestamp(generated_at)

        file_name = "adm_tier_list.txt"

        with open(file_name, 'w', encoding='UTF-8') as file:
            file.write(tier_list)

        if self.controller.write_file(file_name, tier_list):
            await interaction.followup.send(
                file=discord.File(file_name),
                content=f'ADM @ <t:{timestamp}:F>'
            )
            self.controller.delete_file(file_name)

    async def send_system_graph(self, interaction: discord.Interaction, what: str):
        """Send a graph of ADMs matching name"""
        await interaction.response.defer(thinking=True)

        file_name = 'adm_history.png'
        if self.controller.create_history_graph(what, file_name):
            await interaction.followup.send(
                file=discord.File(file_name),
                content=f"ADM History of {what}"
            )
            self.controller.delete_file(file_name)
        else:
            await interaction.followup.send('😥 No ADM data...')

    async def update_adm(self, interaction: discord.Interaction, system_name: str, adm: float):
        """Update ADM for system"""
        await interaction.response.defer(thinking=True)

        if adm <= 0.0 or adm > 6.0:
            await interaction.followup.send(f"Invalid ADM: {adm} (1.0-6.0 valid)")
            return

        if not self.controller.update_system_adm(system_name, adm):
            await interaction.followup.send(f"Couldn't update ADM for {system_name} to {adm}")
        else:
            await interaction.followup.send(f"Manually updated {system_name} ADM to {adm}")

    @app_commands.command(description='Recommend where to raise ADM')
    async def recommend(self, interaction: discord.Interaction):
        """Command recommending which system to raise ADMs"""
        if not check_allowed_channel(interaction.channel, self.configuration.discord['channel']):
            await interaction.response.send_message('Not allowed in this channel.', ephemeral=True)
            return

        await interaction.response.defer()

        recommended_system = self.controller.get_recommended_system()
        if recommended_system is None:
            await interaction.followup.send('😥 No ADM data...')
            return

        system = recommended_system['solarSystemName']
        region = recommended_system['regionName']

        await interaction.followup.send(f'🦀 `{system} in {region}`', tts=True)

    @app_commands.command(description='Posts a summary of all system ADM levels')
    async def summary(self, interaction: discord.Interaction):
        """Command posting a summary of ADMs"""
        if not check_allowed_channel(interaction.channel, self.configuration.discord['channel']):
            await interaction.response.send_message('Not allowed in this channel.', ephemeral=True)
            return

        await self.send_tier_list_summary(interaction)

    @app_commands.command(description='Posts a CSV file of all system ADM levels')
    async def csv(self, interaction: discord.Interaction):
        """Command posting a CSV file of ADMs"""
        if not check_allowed_channel(interaction.channel, self.configuration.discord['channel']):
            await interaction.response.send_message('Not allowed in this channel.', ephemeral=True)
            return

        await interaction.response.defer(thinking=True)

        file_name = "adm_summary.csv"
        if self.controller.create_spreadsheet(file_name):
            await interaction.followup.send(file=discord.File(file_name), content='ADM Spreadsheet')
            self.controller.delete_file(file_name)

    @app_commands.command(description='Post a graph of system ADM.')
    @app_commands.describe(where='The system, constellation, or region to graph.')
    async def history(self, interaction: discord.Interaction, where: str):
        """Command posting a graph of ADM history"""
        if not check_allowed_channel(interaction.channel, self.configuration.discord['channel']):
            await interaction.response.send_message('Not allowed in this channel.', ephemeral=True)
            return

        await self.send_system_graph(interaction, where)

    @app_commands.command(description='Manually refresh all ADM data.')
    async def refresh(self, interaction: discord.Interaction):
        """Command to manually refresh ADM data"""
        if not check_allowed_channel(interaction.channel, self.configuration.discord['channel']):
            await interaction.response.send_message('Not allowed in this channel.', ephemeral=True)
            return

        await interaction.response.defer(thinking=True)
        self.controller.update_adm_data()
        await interaction.followup.send("ADM data manually refreshed 🦀")

    @app_commands.command(description='Manually update ADM of system.')
    @app_commands.describe(system_name='Which system to update ADM')
    async def update(self, interaction: discord.Interaction, system_name: str):
        """Command to manually update ADM for system"""
        if not check_allowed_channel(interaction.channel, self.configuration.discord['channel']):
            await interaction.response.send_message('Not allowed in this channel.', ephemeral=True)
            return

        await interaction.response.send_modal(UpdateAdmModal(system_name, self.controller))

    @commands.Cog.listener()
    async def on_ready(self):
        """cog on_ready event callback"""
        channels = text_channels_with_send_permission(self.bot)

        valid_channels = [
            channel for channel in channels if channel.name == self.configuration.discord['channel']
        ]

        embed = discord.Embed(title=f'🦀 ADM Bot v{pydisadm.__version__}')
        embed.add_field(name='Summary', value='/adm summary')
        embed.add_field(name='CSV', value='/adm csv')
        embed.add_field(name='History', value='/adm history <name>')
        embed.add_field(name='Update', value='/adm update <system>')
        embed.add_field(name='Recommend', value='/adm recommend')
        embed.add_field(name='Refresh', value='/adm refresh')
        embed.add_field(name='Source', value='[github](https://github.com/agelito/adm-bot)')

        for channel in valid_channels:
            await channel.send(embed=embed)

    @commands.command(pass_context=True)
    async def adm_sync_commands(self, ctx: Context):
        """synchronize slash commands"""

        if self.configuration.discord['guild_id']:
            guild = discord.Object(id=self.configuration.discord['guild_id'])
            self.bot.tree.copy_global_to(guild=guild)
            await self.bot.tree.sync(guild=guild)
        else:
            await self.bot.tree.sync()

        await ctx.send('Synchronized bot commands.')



"""pydisadm discord adm bot"""
#!/usr/bin/env python

import signal
import sys
import threading

from dotenv import load_dotenv

from pydisadm.bot.adm_bot import AdmBot
from pydisadm.configuration import Configuration
from pydisadm.controller.adm_controller import AdmController
from pydisadm.loader.static_data import update_static_data
from pydisadm.runnable.runnable_refresh import run_auto_refresh
from pydisadm.services.factory import create_database

def main() -> int:
    """Application main entrypoint"""
    load_dotenv(verbose=True)

    interrupt_event = threading.Event()

    def _signal_handler(sig, _2):
        print(f'Interrupted by signal: {sig}')
        interrupt_event.set()

        sys.exit(0)

    signal.signal(signal.SIGINT, _signal_handler)
    signal.signal(signal.SIGTERM, _signal_handler)

    configuration = Configuration()

    database = create_database(configuration)

    update_static_data(database)

    controller = AdmController(configuration, database)

    controller.update_adm_data()
    controller.purge_adm_records(configuration.database['keep_adm_days'])

    run_auto_refresh(interrupt_event)

    bot = AdmBot(configuration, controller)
    bot.setup_cogs()
    bot.run()

    interrupt_event.set()

    return 0

if __name__ == '__main__':
    sys.exit(main())



"""Common database queries"""

CREATE_TABLE_ADM = """
    CREATE TABLE IF NOT EXISTS adm (
        id INTEGER PRIMARY KEY AUTO_INCREMENT, 
        system_id INTEGER NOT NULL, 
        adm REAL NOT NULL, 
        tier TEXT NOT NULL,
        created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
    )
"""

CREATE_TABLE_ADM_HISTORY = """
    CREATE TABLE IF NOT EXISTS adm_history (
        id INTEGER PRIMARY KEY AUTO_INCREMENT, 
        system_id INTEGER NOT NULL, 
        adm REAL NOT NULL, 
        tier TEXT NOT NULL,
        created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
    )
"""

CREATE_TABLE_MAP = """
    CREATE TABLE IF NOT EXISTS map (
        solarSystemID INTEGER PRIMARY KEY NOT NULL, 
        constellationID INTEGER NOT NULL, 
        regionID INTEGER NOT NULL, 
        solarSystemName TEXT NOT NULL,
        constellationName TEXT NOT NULL,
        regionName TEXT NOT NULL
    )
"""



"""Bot utility functions"""
import discord

def check_allowed_channel(channel, allowed_channel):
    """Check if bot is allowed to respond in channel"""
    if allowed_channel is None:
        return True
    if isinstance(channel, discord.channel.PartialMessageable):
        return True
    if channel.name == allowed_channel:
        return True
    return False

def text_channels_with_send_permission(bot):
    """List of channels with send message permissions"""
    for guild in bot.guilds:
        for channel in guild.channels:
            if (isinstance(channel, discord.TextChannel) and
                channel.permissions_for(channel.guild.me).send_messages):
                yield channel



"""ADM utility functions"""

def adm_from_index(military: int, industrial: int, strategic: int) -> float:
    """Calculate ADM from index values"""
    military_values = [ 0, 0.6, 1.2, 1.7, 2.1, 2.5 ]
    industrial_values = [ 0, 0.6, 1.2, 1.7, 2.1, 2.5 ]
    strategic_values = [ 0, 0.4, 0.6, 0.8, 0.9, 1.0 ]

    adm_sum = sum([
        military_values[military],
        industrial_values[industrial],
        strategic_values[strategic]
    ])

    return round(min(6.0, 1.0 + adm_sum), 2)
